@@ -8,6 +8,8 @@ luatexja.load_module('base');      local ltjb = luatexja.base
 luatexja.load_module('jfont');     local ltjf = luatexja.jfont
 luatexja.load_module('rmlgbm');    local ltjr = luatexja.rmlgbm
 luatexja.load_module('charrange'); local ltjc = luatexja.charrange
+luatexja.load_module('direction'); local ltjd = luatexja.direction
+luatexja.load_module('stack');     local ltjs = luatexja.stack
 
 local id_glyph = node.id('glyph')
 local id_whatsit = node.id('whatsit')
@@ -28,29 +30,36 @@ local to_node = (Dnode ~= node) and Dnode.tonode or nullfunc
 local to_direct = (Dnode ~= node) and Dnode.todirect or nullfunc
 
 local node_new = Dnode.new
-local node_remove = luatexja.Dnode_remove -- Dnode.remove
+local node_remove = Dnode.remove
 local node_next = (Dnode ~= node) and Dnode.getnext or node.next
 local node_free = Dnode.free
 local has_attr = Dnode.has_attribute
 local set_attr = Dnode.set_attribute
 local unset_attr = Dnode.unset_attribute
-local node_insert_after = Dnode.insert_after 
+local node_insert_after = Dnode.insert_after
 local node_write = Dnode.write
 local node_traverse_id = Dnode.traverse_id
 
-local identifiers = fonts.hashes.identifiers
 
 local attr_curjfnt = luatexbase.attributes['ltj@curjfnt']
+local attr_curtfnt = luatexbase.attributes['ltj@curtfnt']
 local attr_yablshift = luatexbase.attributes['ltj@yablshift']
 local attr_ykblshift = luatexbase.attributes['ltj@ykblshift']
+local attr_tablshift = luatexbase.attributes['ltj@tablshift']
+local attr_tkblshift = luatexbase.attributes['ltj@tkblshift']
+local lang_ja = luatexja.lang_ja
+local identifiers = fonts.hashes.identifiers
 
 local ltjf_font_metric_table = ltjf.font_metric_table
+local ltjf_font_extra_info = ltjf.font_extra_info
 local ltjf_find_char_class = ltjf.find_char_class
 local ltjr_cidfont_data = ltjr.cidfont_data
 local ltjc_is_ucs_in_japanese_char = ltjc.is_ucs_in_japanese_char
+local ltjd_get_dir_count = ltjd.get_dir_count
+local dir_tate = luatexja.dir_table.dir_tate
 
 luatexja.userid_table.OTF = luatexbase.newuserwhatsitid('char_by_cid',  'luatexja')
-luatexja.userid_table.VSR = luatexbase.newuserwhatsitid('replace_vs',  'luatexja') 
+luatexja.userid_table.VSR = luatexbase.newuserwhatsitid('replace_vs',  'luatexja')
 local OTF, VSR = luatexja.userid_table.OTF, luatexja.userid_table.VSR
 
 local function get_ucs_from_rmlgbm(c)
@@ -65,7 +74,7 @@ local function get_ucs_from_rmlgbm(c)
       local i = string.len(w)
       if i==4 then -- UCS2
          return tonumber(w,16)
-      elseif i==8 then 
+      elseif i==8 then
          i,w = tonumber(string.sub(w,1,4),16), tonumber(string.sub(w,-4),16)
          if (w>=0xD800) and (w<=0xDB7F) and (i>=0xDC00) and (i<=0xDFFF) then -- Surrogate pair
             return (w-0xD800)*0x400 + (i-0xDC00)
@@ -80,43 +89,49 @@ end
 -- This whatsit node will be extracted to a glyph_node
 local function append_jglyph(char)
    local p = node_new(id_whatsit,sid_user)
-   local v = tex.attribute[attr_curjfnt]
    setfield(p, 'user_id', OTF)
    setfield(p, 'type', 100)
    setfield(p, 'value', char)
-   set_attr(p, attr_yablshift, tex.attribute[attr_ykblshift])
    node_write(p)
 end
 
-local function cid(key)
-   if key==0 then return append_jglyph(char) end
-   local curjfnt = identifiers[tex.attribute[attr_curjfnt]]
-   if not curjfnt.cidinfo or 
-      curjfnt.cidinfo.ordering ~= "Japan1" and
-      curjfnt.cidinfo.ordering ~= "GB1" and
-      curjfnt.cidinfo.ordering ~= "CNS1" and
-      curjfnt.cidinfo.ordering ~= "Korea1" then
---      ltjb.package_warning('luatexja-otf',
---			   'Current Japanese font (or other CJK font) "'
---			      ..curjfnt.psname..'" is not a CID-Keyed font (Adobe-Japan1 etc.)')
-      return append_jglyph(get_ucs_from_rmlgbm(key))
+local cid
+do
+   local tex_get_attr = tex.getattribute
+   cid = function (key)
+      if key==0 then return append_jglyph(char) end
+      local curjfnt = identifiers[tex_get_attr((ltjd_get_dir_count()==dir_tate)
+                                                  and attr_curtfnt or attr_curjfnt)]
+      if not curjfnt.cidinfo or
+         curjfnt.cidinfo.ordering ~= "Japan1" and
+         curjfnt.cidinfo.ordering ~= "GB1" and
+         curjfnt.cidinfo.ordering ~= "CNS1" and
+         curjfnt.cidinfo.ordering ~= "Korea1" then
+         --      ltjb.package_warning('luatexja-otf',
+         --			   'Current Japanese font (or other CJK font) "'
+         --			      ..curjfnt.psname..'" is not a CID-Keyed font (Adobe-Japan1 etc.)')
+            return append_jglyph(get_ucs_from_rmlgbm(key))
+      end
+      local char = curjfnt.resources.unicodes[curjfnt.cidinfo.ordering..'.'..tostring(key)]
+      if not char then
+         ltjb.package_warning('luatexja-otf',
+                              'Current Japanese font (or other CJK font) "'
+                                 ..curjfnt.psname..'" does not have the specified CID character ('
+                                 ..tostring(key)..')',
+                              'Use a font including the specified CID character.')
+         char = 0
+      end
+      return append_jglyph(char)
    end
-   local char = curjfnt.resources.unicodes[curjfnt.cidinfo.ordering..'.'..tostring(key)]
-   if not char then
-      ltjb.package_warning('luatexja-otf',
-                           'Current Japanese font (or other CJK font) "'
-                              ..curjfnt.psname..'" does not have the specified CID character ('
-                              ..tostring(key)..')', 
-                           'Use a font including the specified CID character.')
-      char = 0
-   end
-   return append_jglyph(char)
 end
 
 local function extract(head)
    head = to_direct(head)
    local p = head
-   local v
+   local is_dir_tate = ltjs.list_dir == dir_tate
+   local attr_ablshift = is_dir_tate and attr_tablshift or attr_yablshift
+   local attr_kblshift = is_dir_tate and attr_tkblshift or attr_ykblshift
+   local attr_curfnt =   is_dir_tate and attr_curtfnt or attr_curjfnt
    while p do
       if getid(p)==id_whatsit then
          if getsubtype(p)==sid_user then
@@ -125,14 +140,12 @@ local function extract(head)
                local g = node_new(id_glyph)
                setfield(g, 'subtype', 0)
 	       setfield(g, 'char', getfield(p, 'value'))
-               v = has_attr(p, attr_curjfnt); setfield(g, 'font',v)
-               set_attr(g, attr_curjfnt, puid==OTF and v or -1)
-               -- VSR yields ALchar
-               v = has_attr(p, attr_yablshift)
-               if v then 
-                  set_attr(g, attr_yablshift, v)
+               local v = has_attr(p, attr_curfnt); setfield(g, 'font',v)
+               if puid==OTF then
+                  setfield(g, 'lang', lang_ja)
+                  set_attr(g, attr_kblshift, has_attr(p, attr_kblshift))
                else
-                  unset_attr(g, attr_yablshift)
+                  set_attr(g, attr_ablshift, has_attr(p, attr_ablshift))
                end
                head = node_insert_after(head, p, g)
                head = node_remove(head, p)
@@ -145,16 +158,10 @@ local function extract(head)
    return to_node(head)
 end
 
-luatexbase.add_to_callback('hpack_filter', extract,
-			   'ltj.hpack_filter_otf',
-   luatexbase.priority_in_callback('pre_linebreak_filter',
-				   'ltj.pre_linebreak_filter'))
-luatexbase.add_to_callback('pre_linebreak_filter', extract,
-			   'ltj.pre_linebreak_filter_otf',
-   luatexbase.priority_in_callback('pre_linebreak_filter',
-				   'ltj.pre_linebreak_filter'))
-
-
+ltjb.add_to_callback('hpack_filter', extract,'ltj.otf',
+  luatexbase.priority_in_callback('hpack_filter', 'ltj.main'))
+ltjb.add_to_callback('pre_linebreak_filter', extract,'ltj.otf',
+  luatexbase.priority_in_callback('pre_linebreak_filter', 'ltj.main'))
 -- additional callbacks
 -- 以下は，LuaTeX-ja に用意された callback のサンプルになっている．
 --   JFM の文字クラスの指定の所で，"AJ1-xxx" 形式での指定を可能とした．
@@ -165,20 +172,20 @@ luatexbase.add_to_callback('pre_linebreak_filter', extract,
 local function cid_to_char(fmtable, fn)
    local fi = identifiers[fn]
    if fi.cidinfo and fi.cidinfo.ordering == "Japan1" then
-      fmtable.cid_char_type = {}
       for i, v in pairs(fmtable.chars) do
 	 local j = string.match(i, "^AJ1%-([0-9]*)")
 	 if j then
 	    j = tonumber(fi.resources.unicodes['Japan1.'..tostring(j)])
 	    if j then
-	       fmtable.cid_char_type[j] = v 
+	       fmtable.cid_char_type = fmtable.cid_char_type  or {}
+	       fmtable.cid_char_type[j] = v
 	    end
 	 end
       end
    end
    return fmtable
 end
-luatexbase.add_to_callback("luatexja.define_jfont", 
+luatexbase.add_to_callback("luatexja.define_jfont",
 			   cid_to_char, "ltj.otf.define_jfont", 1)
 --  既に読み込まれているフォントに対しても，同じことをやらないといけない
 for fn, v in pairs(ltjf_font_metric_table) do
@@ -193,134 +200,57 @@ local function cid_set_char_class(arg, fmtable, char)
    else return 0
    end
 end
-luatexbase.add_to_callback("luatexja.find_char_class", 
+luatexbase.add_to_callback("luatexja.find_char_class",
 			   cid_set_char_class, "ltj.otf.find_char_class", 1)
 
 -------------------- IVS
-local font_ivs_table = {} -- key: fontnumber
-local enable_ivs
+local enable_ivs, disable_ivs
 do
    local is_ivs_enabled = false
-   local ivs -- temp table
-   local sort = table.sort
-   local uniq_flag
-   local function add_ivs_table(tg, unitable, glyphmax)
-      for i = 0, glyphmax-1 do
-	 if tg[i] then
-	    local gv = tg[i]
-	    if gv.altuni then
-	       for _,at in pairs(gv.altuni) do
-		  local bu, vsel = at.unicode, at.variant
-		  if vsel then
-		     if vsel>=0xE0100 then vsel = vsel - 0xE0100 end
-		     if not ivs[bu] then ivs[bu] = {} end
-		     uniq_flag = true
-		     for i,_ in pairs(ivs[bu]) do
-			if i==vs then uniq_flag = false; break end
-		     end
-		     if uniq_flag then 
-			ivs[bu][vsel] = unitable[gv.name]
-		     end
-		  end
-	       end
-	    end
-	 end
-      end
-   end
-   local function make_ivs_table(id, fname)
-      ivs = {}
-      local fl = fontloader.open(fname)
-      local unicodes = id.resources.unicodes
-      if fl.glyphs then
-	 add_ivs_table(fl.glyphs, id.resources.unicodes, fl.glyphmax)
-      end
-      if fl.subfonts then
-         for _,v in pairs(fl.subfonts) do
-            add_ivs_table(v.glyphs, id.resources.unicodes, v.glyphmax)
-         end
-      end
-      fontloader.close(fl)
-      return ivs
-   end
-
--- loading and saving
-   local font_ivs_basename = {} -- key: basename
-   local cache_ver = 4
-   local checksum = file.checksum
-
-   local function prepare_ivs_data(n, id)
-      -- test if already loaded
-      if type(id)=='number' then -- sometimes id is an integer
-         font_ivs_table[n] = font_ivs_table[id]; return
-      elseif not id then return
-      end
-      local fname = id.filename
-      local bname = file.basename(fname)
-      if not fname then 
-         font_ivs_table[n] = {}; return
-      elseif font_ivs_basename[bname] then 
-         font_ivs_table[n] = font_ivs_basename[bname]; return
-      end
-      
-      -- if the cache is present, read it
-      local newsum = checksum(fname) -- MD5 checksum of the fontfile
-      local v = "ivs_" .. string.lower(file.nameonly(fname))
-      local dat = ltjb.load_cache(v, 
-         function (t) return (t.version~=cache_ver) or (t.chksum~=newsum) end
-      )
-      -- if the cache is not found or outdated, save the cache
-      if dat then 
-	 font_ivs_basename[bname] = dat[1] or {}
-      else
-	 dat = make_ivs_table(id, fname)
-	 font_ivs_basename[bname] = dat or {}
-	 ltjb.save_cache( v,
-			  {
-			     chksum = checksum(fname), 
-			     version = cache_ver,
-			     dat,
-			  })
-      end
-      font_ivs_table[n] = font_ivs_basename[bname]
-   end
-
 -- 組版時
    local function ivs_jglyph(char, bp, pf, uid)
       local p = node_new(id_whatsit,sid_user)
       setfield(p, 'user_id', uid)
       setfield(p, 'type', 100)
       setfield(p, 'value', char)
-      set_attr(p, attr_curjfnt, pf)
-      set_attr(p, attr_yablshift, has_attr(bp, attr_ykblshift) or 0)
       return p
    end
 
-   local function do_ivs_repr(head)
-      head = to_direct(head)
+   local function do_ivs_repr(h)
+      local head = to_direct(h)
       local p, r = head
+      local is_dir_tate = (ltjs.list_dir == dir_tate)
+      local attr_ablshift = is_dir_tate and attr_tablshift or attr_yablshift
+      local attr_kblshift = is_dir_tate and attr_tkblshift or attr_ykblshift
+      local attr_curfnt =   is_dir_tate and attr_curtfnt or attr_curjfnt
       while p do
 	 local pid = getid(p)
 	 if pid==id_glyph then
-            local pf = getfont(p)
             local q = node_next(p) -- the next node of p
             if q and getid(q)==id_glyph then
                local qc = getchar(q)
-               if (qc>=0xFE00 and qc<=0xFE0F) or (qc>=0xE0100 and qc<0xE01F0) then 
-                  -- q is a variation selector
+               if (qc>=0xFE00 and qc<=0xFE0F) or (qc>=0xE0100 and qc<0xE01F0) then
+		   -- q is a variation selector
                   if qc>=0xE0100 then qc = qc - 0xE0100 end
-                  local pt = font_ivs_table[pf]
+                  local pf = getfont(p)
+                  local pt = ltjf_font_extra_info[pf]
                   pt = pt and pt[getchar(p)];  pt = pt and  pt[qc]
                   head, r = node_remove(head,q)
 		  node_free(q)
                   if pt then
+		     local is_jachar = (getfield(p, 'lang')==lang_ja)
                      local np = ivs_jglyph(pt, p, pf,
-                                           (has_attr(p,attr_curjfnt) or 0)==pf and OTF or VSR)
-                     head = node_insert_after(head, p, np) 
+                                           is_jachar and OTF or VSR)
+		     if is_jachar then
+			set_attr(np, attr_curfnt, pf)
+			set_attr(np, attr_kblshift, has_attr(p, attr_kblshift))
+		     end
+                     head = node_insert_after(head, p, np)
                      head = node_remove(head,p)
 		     node_free(p)
-                  end
+		  end
 		  p = r
-	       else 
+	       else
 		  p = q
                end
 	    else
@@ -329,15 +259,8 @@ do
 	 else
 	    p = node_next(p)
          end
-      end
-      return to_node(head)
-   end
-
-   -- font define
-   local function font_callback(name, size, id, fallback)
-      local d = fallback(name, size, id)
-      prepare_ivs_data(id, d)
-      return d
+     end
+     return to_node(head)
    end
 
    enable_ivs = function ()
@@ -345,29 +268,26 @@ do
 	 ltjb.package_warning('luatexja-otf',
 			      'luatexja.otf.enable_ivs() was already called, so this call is ignored', '')
       else
-	 luatexbase.add_to_callback('hpack_filter', 
-				    do_ivs_repr,'do_ivs', 1)
-	 luatexbase.add_to_callback('pre_linebreak_filter', 
-				    do_ivs_repr, 'do_ivs', 1)
-	 local ivs_callback = function (name, size, id)
-	    return font_callback(
-	       name, size, id, 
-	       function (name, size, id) return luatexja.font_callback(name, size, id) end
-	    )
-	 end
-	 luatexbase.add_to_callback('define_font',ivs_callback,"luatexja.ivs_font_callback", 1)
-	 for i=1,font.nextid()-1 do
-	    if identifiers[i] then prepare_ivs_data(i, identifiers[i]) end
-	 end
+	 ltjb.add_to_callback('hpack_filter', do_ivs_repr, 'ltj.do_ivs',
+            luatexbase.priority_in_callback('hpack_filter', 'luaotfload.node_processor'))
+	 ltjb.add_to_callback('pre_linebreak_filter', do_ivs_repr, 'ltj.do_ivs',
+            luatexbase.priority_in_callback('pre_linebreak_filter', 'luaotfload.node_processor'))
 	 is_ivs_enabled = true
       end
    end
+   disable_ivs = function ()
+      if is_ivs_enabled then
+	 luatexbase.remove_from_callback('hpack_filter', 'ltj.do_ivs')
+	 luatexbase.remove_from_callback('pre_linebreak_filter', 'ltj.do_ivs')
+	 is_ivs_enabled = false
+      end
+   end
 end
 
 luatexja.otf = {
   append_jglyph = append_jglyph,
   enable_ivs = enable_ivs,  -- 隠し機能: IVS
-  font_ivs_table = font_ivs_table,
+  disable_ivs = disable_ivs,  -- 隠し機能: IVS
   cid = cid,
 }
 
