@@ -1,5 +1,6 @@
 
 require('lualibs')
+tableunpack = table.unpack
 
 ------------------------------------------------------------------------
 -- naming:
@@ -13,12 +14,22 @@ function luatexja.load_lua(fn)
    local found = kpse.find_file(fn, 'tex')
    if not found then
       tex.error("LuaTeX-ja error: File `" .. fn .. "' not found")
-   else 
+   else
       texio.write_nl('(' .. found .. ')')
       dofile(found)
    end
 end
 
+-- check token library
+if newtoken then
+   luatexja.token = newtoken
+else
+   local luatex_version = status.list().luatex_version
+   if luatex_version >80 then
+      luatexja.token = token
+   end
+end
+
 --- 以下は全ファイルで共有される定数
 local icflag_table = {}
 luatexja.icflag_table = icflag_table
@@ -58,65 +69,45 @@ luatexja.userid_table = userid_table
 userid_table.IHB  = luatexbase.newuserwhatsitid('inhibitglue',  'luatexja') -- \inhibitglue
 userid_table.STCK = luatexbase.newuserwhatsitid('stack_marker', 'luatexja') -- スタック管理
 userid_table.BPAR = luatexbase.newuserwhatsitid('begin_par',    'luatexja') -- 「段落始め」
+userid_table.DIR  = luatexbase.newuserwhatsitid('direction',    'luatexja') -- 組方向
 
-------------------------------------------------------------------------
--- FIX node.remove
-------------------------------------------------------------------------
-do
-   local node_remove, node_next, node_prev = node.remove, node.next, node.prev
-   function luatexja.node_remove (head, current)
-      if head==current then
-         local q, r = node_next(current), node_prev(current)
-         if q then q.prev = r end
-         if r and node_next(r)==current then 
-            r.next = q
-         end
-         return q, q
-      else
-         return node_remove(head, current)
-      end
-   end
-   local Dnode = node.direct or node
-   if Dnode~=node then
-      local Dnode_remove, setfield = Dnode.remove, Dnode.setfield
-      local Dnode_next, Dnode_prev = Dnode.getnext, Dnode.getprev
-      function luatexja.Dnode_remove (head, current)
-         if head==current then
-            local q, r = Dnode_next(current), Dnode_prev(current)
-            if q then setfield(q, 'prev', r) end
-            if r and Dnode_next(r) == current then
-               setfield(r, 'next', q)
-            end
-            return q, q
-         else
-            return Dnode_remove(head, current)
-         end
-      end
-   else
-      luatexja.Dnode_remove = luatexja.node_remove 
-   end
-end
-
+local dir_table = {}
+luatexja.dir_table = dir_table
+dir_table.dir_dtou = 1
+dir_table.dir_tate = 3
+dir_table.dir_yoko = 4
+dir_table.dir_math_mod    = 8
+dir_table.dir_node_auto   = 128 -- 組方向を合わせるために自動で作られたもの
+dir_table.dir_node_manual = 256 -- 寸法代入によって作られたもの
+dir_table.dir_utod = dir_table.dir_tate + dir_table.dir_math_mod
+   -- 「縦数式ディレクション」 in pTeX
 --- 定義終わり
 
 local load_module = luatexja.load_module
 load_module('base');      local ltjb = luatexja.base
 load_module('rmlgbm');    local ltjr = luatexja.rmlgbm -- must be 1st
+
+if luatexja_debug then load_module('debug') end
+
 load_module('charrange'); local ltjc = luatexja.charrange
+load_module('stack');     local ltjs = luatexja.stack
+load_module('direction'); local ltjd = luatexja.direction -- +1 hlist +1 attr_list
 load_module('jfont');     local ltjf = luatexja.jfont
 load_module('inputbuf');  local ltji = luatexja.inputbuf
-load_module('stack');     local ltjs = luatexja.stack
 load_module('pretreat');  local ltjp = luatexja.pretreat
-load_module('jfmglue');   local ltjj = luatexja.jfmglue
 load_module('setwidth');  local ltjw = luatexja.setwidth
+load_module('jfmglue');   local ltjj = luatexja.jfmglue -- +1 glue +1 gs +1 attr_list
 load_module('math');      local ltjm = luatexja.math
 load_module('tangle');    local ltjb = luatexja.base
 
+
 local attr_jchar_class = luatexbase.attributes['ltj@charclass']
+local attr_jchar_code = luatexbase.attributes['ltj@charcode']
 local attr_curjfnt = luatexbase.attributes['ltj@curjfnt']
 local attr_yablshift = luatexbase.attributes['ltj@yablshift']
 local attr_icflag = luatexbase.attributes['ltj@icflag']
 local attr_uniqid = luatexbase.attributes['ltj@uniqid']
+local attr_dir = luatexbase.attributes['ltj@dir']
 local cat_lp = luatexbase.catcodetables['latex-package']
 
 -- Three aux. functions, bollowed from tex.web
@@ -127,14 +118,14 @@ local floor = math.floor
 local function print_scaled(s)
    local out=''
    local delta=10
-   if s<0 then 
+   if s<0 then
       out=out..'-'; s=-s
    end
    out=out..tostring(floor(s/unity)) .. '.'
    s=10*(s%unity)+5
    repeat
       if delta>unity then s=s+32768-50000 end
-      out=out .. tostring(floor(s/unity)) 
+      out=out .. tostring(floor(s/unity))
       s=10*(s%unity)
       delta=delta*10
    until s<=delta
@@ -148,7 +139,7 @@ local function print_glue(d,order)
       while order>1 do
 	 out=out..'l'; order=order-1
       end
-   else 
+   else
       out=out..'pt'
    end
    return out
@@ -167,22 +158,28 @@ end
 
 
 ------------------------------------------------------------------------
--- CODE FOR GETTING/SETTING PARAMETERS 
+-- CODE FOR GETTING/SETTING PARAMETERS
 ------------------------------------------------------------------------
 
 -- EXT: print parameters that don't need arguments
 do
    luatexja.unary_pars = {
-      yalbaselineshift = function(t) 
+      yalbaselineshift = function(t)
 	 return print_scaled(tex.getattribute('ltj@yablshift'))..'pt'
       end,
-      yjabaselineshift = function(t) 
+      yjabaselineshift = function(t)
 	 return print_scaled(tex.getattribute('ltj@ykblshift'))..'pt'
       end,
-      kanjiskip = function(t) 
+      talbaselineshift = function(t)
+	 return print_scaled(tex.getattribute('ltj@tablshift'))..'pt'
+      end,
+      tjabaselineshift = function(t)
+	 return print_scaled(tex.getattribute('ltj@tkblshift'))..'pt'
+      end,
+      kanjiskip = function(t)
 	 return print_spec(ltjs.get_stack_skip(stack_table_index.KSK, t))
       end,
-      xkanjiskip = function(t) 
+      xkanjiskip = function(t)
 	 return print_spec(ltjs.get_stack_skip(stack_table_index.XSK, t))
       end,
       jcharwidowpenalty = function(t)
@@ -205,7 +202,15 @@ do
 	 elseif f == math.two_add then r = 'both'
 	 end
 	 return r
-      end
+      end,
+      direction = function()
+	 local v = ltjd.get_dir_count()
+	 if math.abs(tex.nest[tex.nest.ptr].mode) == ltjs.mmode and v == dir_table.dir_tate then
+	    v = dir_table.dir_utod
+	 end
+	 return v
+      end,
+      adjustdir = ltjd.get_adjust_dir_count,
    }
 
    local unary_pars = luatexja.unary_pars
@@ -213,6 +218,7 @@ do
       if unary_pars[k] then
 	 tex.write(tostring(unary_pars[k](tex.getcount('ltj@@stack'))))
       end
+      ltjb.stop_time_measure('get_par')
    end
 end
 
@@ -235,15 +241,15 @@ do
 	 return (c<0) and 1 or ltjc.get_range_setting(c)
       end,
       prebreakpenalty = function(c, t)
-	 return ltjs.get_stack_table(stack_table_index.PRE 
+	 return ltjs.get_stack_table(stack_table_index.PRE
 					  + ltjb.in_unicode(c, true), 0, t)
       end,
       postbreakpenalty = function(c, t)
-	 return ltjs.get_stack_table(stack_table_index.POST 
+	 return ltjs.get_stack_table(stack_table_index.POST
 					  + ltjb.in_unicode(c, true), 0, t)
       end,
       kcatcode = function(c, t)
-	 return ltjs.get_stack_table(stack_table_index.KCAT 
+	 return ltjs.get_stack_table(stack_table_index.KCAT
 					  + ltjb.in_unicode(c, false), 0, t)
       end,
       chartorange = function(c, t)
@@ -253,14 +259,25 @@ do
 	 return ltjs.get_stack_table(stack_table_index.XSP
 					  + ltjb.in_unicode(c, true), 3, t)
       end,
+      boxdir = function(c, t)
+	 if type(c)~='number' or c<0 or c>65535 then
+            ltjb.package_error('luatexja',
+                               'Bad register code (' .. tostring(c) .. ')',
+                               'A register must be between 0 and 65535.\n'..
+                                  'I changed this one to zero.')
+            c=0
+         end
+         return ltjd.get_register_dir(c)
+      end,
    }
-   local binary_pars = luatexja.binary_pars 
+   local binary_pars = luatexja.binary_pars
 
    binary_pars.alxspmode = binary_pars.jaxspmode
    function luatexja.ext_get_parameter_binary(k,c)
       if binary_pars[k] then
 	 tex.write(tostring(binary_pars[k](c,tex.getcount('ltj@@stack'))))
       end
+      ltjb.stop_time_measure('get_par')
    end
 end
 
@@ -269,65 +286,62 @@ function luatexja.ext_print_global()
    if luatexja.isglobal=='global' then tex.sprint(cat_lp, '\\global') end
 end
 
+
 -- main process
 do
+   local start_time_measure, stop_time_measure
+      = ltjb.start_time_measure, ltjb.stop_time_measure
    local Dnode = node.direct or node
    local nullfunc = function (n) return n end
    local to_node = (Dnode ~= node) and Dnode.tonode or nullfunc
    local to_direct = (Dnode ~= node) and Dnode.todirect or nullfunc
+   local ensure_tex_attr = ltjb.ensure_tex_attr
+
    -- mode = true iff main_process is called from pre_linebreak_filter
-   local function main_process(head, mode, dir)
-      tex.setattribute('global', attr_icflag, 0)
-      local p = to_direct(head)
-      p = ltjj.main(p,mode)
-      if p then p = ltjw.set_ja_width(p, dir) end
-      return to_node(p)
+   local function main_process(head, mode, dir, gc)
+      ensure_tex_attr(attr_icflag, 0)
+      if gc == 'fin_row' then return head
+      else
+            --luatexja.ext_show_node_list(head, 'T> ', print)
+	    start_time_measure('jfmglue')
+	    local p = ltjj.main(to_direct(head),mode, dir)
+	    stop_time_measure('jfmglue')
+	    return to_node(p)
+      end
    end
 
    local function adjust_icflag(h)
       -- kern from luaotfload will have icflag = 1
       -- (same as italic correction)
-      tex.setattribute('global', attr_icflag, 1)
+      ensure_tex_attr(attr_icflag, 1)
       return h
    end
 
    -- callbacks
-   
-   luatexbase.add_to_callback(
-      'pre_linebreak_filter', 
+   ltjb.add_to_callback(
+      'pre_linebreak_filter',
       function (head,groupcode)
-	 return main_process(head, true, tex.textdir)
-      end,'ltj.pre_linebreak_filter',
-      luatexbase.priority_in_callback('pre_linebreak_filter',
-				      'luaotfload.node_processor') + 1)
-   luatexbase.add_to_callback(
-      'hpack_filter', 
+	 return main_process(head, true, tex.textdir, groupcode)
+      end,'ltj.main',
+      luatexbase.priority_in_callback('pre_linebreak_filter', 'luaotfload.node_processor')+1)
+   ltjb.add_to_callback(
+      'hpack_filter',
       function (head,groupcode,size,packtype, dir)
-	 return main_process(head, false, dir)
-      end,'ltj.hpack_filter',
-      luatexbase.priority_in_callback('hpack_filter',
-				      'luaotfload.node_processor') + 1)
-   luatexbase.add_to_callback('pre_linebreak_filter', adjust_icflag, 'adjust_icflag', 1)
-   luatexbase.add_to_callback('hpack_filter', adjust_icflag, 'adjust_icflag', 1)
+	 return main_process(head, false, dir, groupcode)
+      end,'ltj.main',
+      luatexbase.priority_in_callback('hpack_filter', 'luaotfload.node_processor')+1)
+   ltjb.add_to_callback('pre_linebreak_filter', adjust_icflag, 'ltj.adjust_icflag', 1)
+   ltjb.add_to_callback('hpack_filter', adjust_icflag, 'ltj.adjust_icflag', 1)
 
 end
 
--- define_font
-do
-   local otfl_fdr = fonts.definers.read
-   local ltjr_font_callback = ltjr.font_callback
-   function luatexja.font_callback(name, size, id)
-      return ltjf.font_callback(
-	 name, size, id, 
-	 function (name, size, id) return ltjr_font_callback(name, size, id, otfl_fdr) end
-      )
-   end
-   luatexbase.add_to_callback('define_font',luatexja.font_callback,"luatexja.font_callback", 1)
+-- cleanup process
+function luatexja.ext_cleanup()
+   ltjf.cleanup_size_cache()
+   ltjd.remove_end_whatsit()
 end
 
 
-
-
 -- debug
 
 do
@@ -352,32 +366,40 @@ local function get_attr_icflag(p)
    return (has_attr(p, attr_icflag) or 0) % icflag_table.PROCESSED_BEGIN_FLAG
 end
 
-local debug_depth
+local prefix, inner_depth
 
-local function debug_show_node_X(p,print_fn)
-   local k = debug_depth
+local function debug_show_node_X(p,print_fn, limit)
+   local k = prefix
    local s
    local pt=node_type(p.id)
-   local base = debug_depth .. string.format('%X', get_attr_icflag(p))
+   local base = prefix .. string.format('%X', get_attr_icflag(p))
    .. ' ' .. pt .. ' ' .. tostring(p.subtype) .. ' '
    if pt == 'glyph' then
-      s = base .. ' ' .. utf.char(p.char) .. ' '  .. tostring(p.font)
-         .. ' (' .. print_scaled(p.height) .. '+' 
+      s = base .. ' ' .. utf.char(p.char) .. ' '
+         .. tostring(p.font)
+         .. ' (' .. print_scaled(p.height) .. '+'
          .. print_scaled(p.depth) .. ')x' .. print_scaled(p.width)
       print_fn(s)
-   elseif pt=='hlist' or pt=='vlist' then
-      s = base .. '(' .. print_scaled(p.height) .. '+' 
-         .. print_scaled(p.depth) .. ')x' .. print_scaled(p.width) .. p.dir
-      if p.shift~=0 then
+   elseif pt=='hlist' or pt=='vlist' or pt=='unset'or pt=='ins' then
+      if pt=='ins' then
+	 s = base .. '(' .. print_scaled(p.height) .. '+'
+	    .. print_scaled(p.depth) .. ')'
+	    .. ', dir=' .. tostring(node.has_attribute(p, attr_dir))
+      else
+	 s = base .. '(' .. print_scaled(p.height) .. '+'
+	    .. print_scaled(p.depth) .. ')x' .. print_scaled(p.width)
+	    .. ', dir=' .. tostring(node.has_attribute(p, attr_dir))
+      end
+      if (p.shift or 0)~=0 then
          s = s .. ', shifted ' .. print_scaled(p.shift)
       end
-      if p.glue_sign >= 1 then 
+      if p.glue_set and p.glue_sign ==2 or ( p.glue_sign==1 and p.glue_set>0) then
          s = s .. ' glue set '
          if p.glue_sign == 2 then s = s .. '-' end
          s = s .. tostring(floor(p.glue_set*10000)/10000)
-         if p.glue_order == 0 then 
-            s = s .. 'pt' 
-         else 
+         if p.glue_order == 0 then
+            s = s .. 'pt'
+         else
             s = s .. 'fi'
             for i = 2,  p.glue_order do s = s .. 'l' end
          end
@@ -385,14 +407,23 @@ local function debug_show_node_X(p,print_fn)
       if get_attr_icflag(p) == icflag_table.PACKED then
          s = s .. ' (packed)'
       end
-      print_fn(s); debug_depth=debug_depth.. '.'
-      for q in node.traverse(p.head) do
-         debug_show_node_X(q, print_fn)
+      print_fn(s);
+      local bid = inner_depth
+      prefix, inner_depth = prefix.. '.', inner_depth + 1
+      if inner_depth < limit then
+	 for q in node.traverse(p.head) do
+	    debug_show_node_X(q, print_fn, limit)
+	 end
       end
-      debug_depth=k
+      prefix=k
+   elseif pt=='rule' then
+      s = base .. '(' .. print_scaled(p.height) .. '+'
+         .. print_scaled(p.depth) .. ')x' .. print_scaled(p.width)
+	 .. ', dir=' .. tostring(node.has_attribute(p, attr_dir))
+      print_fn(s)
    elseif pt == 'glue' then
       s = base .. ' ' ..  print_spec(p.spec)
-      if get_attr_icflag(p)>icflag_table.KINSOKU 
+      if get_attr_icflag(p)>icflag_table.KINSOKU
          and get_attr_icflag(p)<icflag_table.KANJI_SKIP then
          s = s .. ' (from JFM: priority ' .. get_attr_icflag(p)-icflag_table.FROM_JFM .. ')'
       elseif get_attr_icflag(p)==icflag_table.KANJI_SKIP then
@@ -413,7 +444,7 @@ local function debug_show_node_X(p,print_fn)
 	 s = s .. ' (italic correction)'
          -- elseif get_attr_icflag(p)==ITALIC then
          --    s = s .. ' (italic correction)'
-      elseif get_attr_icflag(p)>icflag_table.KINSOKU 
+      elseif get_attr_icflag(p)>icflag_table.KINSOKU
          and get_attr_icflag(p)<icflag_table.KANJI_SKIP then
 	 s = s .. ' (from JFM: priority ' .. get_attr_icflag(p)-icflag_table.FROM_JFM .. ')'
       end
@@ -425,44 +456,60 @@ local function debug_show_node_X(p,print_fn)
       end
       print_fn(s)
    elseif pt == 'whatsit' then
-      s = base .. ' subtype: ' ..  tostring(p.subtype)
+      s = base
       if p.subtype==sid_user then
-         if p.type ~= 110 then 
-            s = s .. ' user_id: ' .. p.user_id .. ' ' .. p.value
+	 local t = tostring(p.user_id) .. ' (' ..
+	    luatexbase.get_user_whatsit_name(p.user_id) .. ') '
+         if p.type ~= 110 then
+            s = s .. ' userid:' .. t .. p.value
             print_fn(s)
          else
-            s = s .. ' user_id: ' .. p.user_id .. ' (node list)'
+            s = s .. ' userid:' .. t .. '(node list)'
+	    if p.user_id==userid_table.DIR then
+	       s = s .. ' dir: ' .. tostring(node.has_attribute(p, attr_dir))
+	    end
             print_fn(s)
-            local q = p.value
-            debug_depth=debug_depth.. '.'
-            while q do 
-               debug_show_node_X(q, print_fn); q = node_next(q)
-            end
-            debug_depth=k
+	    local bid = inner_depth
+            prefix, inner_depth =prefix.. '.', inner_depth + 1
+            if inner_depth < limit then
+	       for q in node.traverse(p.value) do
+		  debug_show_node_X(q, print_fn, limit)
+	       end
+	    end
+            prefix, inner_depth = k, bid
          end
       else
-         s = s .. node.subtype(p.subtype); print_fn(s)
+         s = s .. (node.subtype(p.subtype) or '')
+	 if p.subtype==1 then
+	    s = s .. ' stream=' .. p.stream
+	    print_fn(s)
+	    for i=1,#p.data do
+	       print_fn(s .. '  [' .. i .. '] = ' .. tostring(p.data[i].csname))
+	    end
+	 else
+	    print_fn(s)
+	 end
       end
    -------- math node --------
    elseif pt=='noad' then
       s = base ; print_fn(s)
       if p.nucleus then
-         debug_depth = k .. 'N'; debug_show_node_X(p.nucleus, print_fn); 
+         prefix = k .. 'N'; debug_show_node_X(p.nucleus, print_fn);
       end
       if p.sup then
-         debug_depth = k .. '^'; debug_show_node_X(p.sup, print_fn); 
+         prefix = k .. '^'; debug_show_node_X(p.sup, print_fn);
       end
       if p.sub then
-         debug_depth = k .. '_'; debug_show_node_X(p.sub, print_fn); 
+         prefix = k .. '_'; debug_show_node_X(p.sub, print_fn);
       end
-      debug_depth = k;
+      prefix = k;
    elseif pt=='math_char' then
       s = base .. ' fam: ' .. p.fam .. ' , char = ' .. utf.char(p.char)
       print_fn(s)
    elseif pt=='sub_box' or pt=='sub_mlist' then
       print_fn(base)
       if p.head then
-         debug_depth = k .. '.'; 
+         prefix = k .. '.';
 	 for q in node.traverse(p.head) do
 	    debug_show_node_X(q, print_fn)
 	 end
@@ -472,22 +519,24 @@ local function debug_show_node_X(p,print_fn)
    end
    p=node_next(p)
 end
-function luatexja.ext_show_node_list(head,depth,print_fn)
-   debug_depth = depth
+function luatexja.ext_show_node_list(head,depth,print_fn, lim)
+   prefix = depth
+   inner_depth = 0
    if head then
       while head do
-         debug_show_node_X(head, print_fn); head = node_next(head)
+         debug_show_node_X(head, print_fn, lim or 1/0); head = node_next(head)
       end
    else
-      print_fn(debug_depth .. ' (null list)')
+      print_fn(prefix .. ' (null list)')
    end
 end
-function luatexja.ext_show_node(head,depth,print_fn)
-   debug_depth = depth
+function luatexja.ext_show_node(head,depth,print_fn, lim)
+   prefix = depth
+   inner_depth = 0
    if head then
-      debug_show_node_X(head, print_fn)
+      debug_show_node_X(head, print_fn, lim or 1/0)
    else
-      print_fn(debug_depth .. ' (null list)')
+      print_fn(prefix .. ' (null list)')
    end
 end
 
